/*
 * utf8 ru
  *      Author: alexrayne <alexraynepe196@gmail.com>
  ------------------------------------------------------------------------
    Copyright (c) alexrayne

   All rights reserved.
   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:
   - Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   - Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   - Neither the name of ARM nor the names of its contributors may be used
     to endorse or promote products derived from this software without
     specific prior written permission.
   *
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDERS AND CONTRIBUTORS BE
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
   POSSIBILITY OF SUCH DAMAGE. *
  ------------------------------------------------------------------------
 * */

#ifndef INTFACES_H
#define INTFACES_H

#include <stdarg.h>
#include "hal_types.h"

#ifdef __unix__
    #include <assert.h>
#else
#include <assert.h>
#endif


#if 1
#define HAL_ASSERT(...) assert(__VA_ARGS__)
#else
#define HAL_ASSERT(...)
#endif



#define ERROR_MAX_LENGTH    25



class HAL_Device{
public:
    bool init_state;
    typedef const_string dev_name;
    typedef const_string error_str;

    HAL_Device(const_string name):init_state(false), _name(name), error(NULL){};
    virtual int init() = 0;
    virtual int deinit() = 0;

    virtual const_string getLastErrorStr(){
      return this->error;
    };

    const_string name() const {return this->_name;};

protected:
    dev_name    _name;
    error_str   error;
};

class Power_Manager{
public:
    enum POWER_STATE{
        WORK,
        SLEEP,
        HALT
    };

    virtual int setPower(POWER_STATE state){return DEV_NOT_IMPLEMENTED;}
    virtual POWER_STATE getPower(){return this->state;}

protected:
    POWER_STATE state;//TODO = HALT;
};




class HAL_IO_Device{
public:
    typedef const_string ioctl_string;
    virtual int IOCtrl(ioctl_string cmd){return DEV_NOT_IMPLEMENTED;}
    virtual DevResult readByte(u8 *data){return DEV_NOT_IMPLEMENTED;}
    virtual DevResult writeByte(u8 data){return DEV_NOT_IMPLEMENTED;}
    virtual int readData(u8 *data, size_t size){return DEV_NOT_IMPLEMENTED;}
    virtual int writeData(const u8 *data, size_t size){return DEV_NOT_IMPLEMENTED;}

private:

};

class HAL_ASYNCIO_Device{
public:
    typedef const_string ioctl_string;
    virtual int IOCtrl(ioctl_string cmd){return DEV_NOT_IMPLEMENTED;}

    // \return - bytes send amount
    virtual int postByte(u8 data){return DEV_NOT_IMPLEMENTED;}
    virtual int postData(const void *data, size_t size){return DEV_NOT_IMPLEMENTED;}

    static const unsigned toInfinite = ~0u;
    virtual PTResult wait_flush(unsigned to = toInfinite){
        return AS_PTRESULT(DEV_NOT_IMPLEMENTED);
    };
    // \return - bytes avail for read
    virtual PTResult wait_avail(size_t size, unsigned to = toInfinite){
        return AS_PTRESULT(DEV_NOT_IMPLEMENTED);
    };
};

class HAL_IOAdress_Device
//        :public HAL_IO_Device
{
    public:
        typedef u32 ioaddr_t;
        ioaddr_t    io_addr() const {return ioaddr;};
        virtual int io_addr(ioaddr_t x) {ioaddr = x; return x;};
    protected:
        ioaddr_t    ioaddr;
};

class HAL_REGs_Device{
public:
    typedef const_string ioctl_string;
    typedef u32 regaddr_t;
    virtual int IOCtrl(ioctl_string cmd){return DEV_NOT_IMPLEMENTED;}
    virtual DevResult readByte(regaddr_t address, u8 *data){return DEV_NOT_IMPLEMENTED;}
    virtual DevResult writeByte(regaddr_t address, u8 data){return DEV_NOT_IMPLEMENTED;}
    virtual int readData(regaddr_t address, void *data, size_t size){return DEV_NOT_IMPLEMENTED;}
    virtual int writeData(regaddr_t address, const void *data, size_t size){return DEV_NOT_IMPLEMENTED;}

private:

};

class HAL_IO_Block_Device{
public:
    typedef const_string ioctl_string;
    typedef uint32_t addr_t;
    virtual int IOCtrl(const_string cmd){return DEV_NOT_IMPLEMENTED;}
    virtual DevResult readByte(addr_t address, u8 *data){return DEV_NOT_IMPLEMENTED;}
    virtual DevResult writeByte(addr_t address, u8 data){return DEV_NOT_IMPLEMENTED;}
    virtual int readData(addr_t address, void *data, size_t size){return DEV_NOT_IMPLEMENTED;}
    virtual int writeData(addr_t address, const void *data, size_t size){return DEV_NOT_IMPLEMENTED;}

private:

};

class HAL_OUTBlock_Device{
public:
  //* блокирующая печать
  virtual int puts( const char* str) = 0;
  //*  \return - длинна отправленного участка
  virtual int putData ( const void* str, unsigned len) = 0;

  //*  ожидание доступности печати
  //*  \return - количество байт возможных для неблокирующей печати
  virtual int put_wait(unsigned to = 0) = 0;
  //*  почти тоже put_wait, ждет полного опустошения
  virtual int put_flush(unsigned to = 0) = 0;
  //*  очищает буфер, прерывая текущую отправку
  virtual int put_drop() = 0;

  //* монополизация вывода (puts, putData предпочтительно использую ее )
  //* \arg onoff - захват/освобождение
  //* \return    - состояние захвачн ли вывод
  static const unsigned toInfinite = ~0;
  virtual bool put_access(bool onoff, unsigned to = toInfinite) = 0;
};

class StdOUT_Device
    : public HAL_OUTBlock_Device
{
public:
    //* блокирующая печать
    virtual int putChar(int ch) = 0;

    //*  неблокирующая печать
    virtual int postData ( const void* str, unsigned len) = 0;

};

class HAL_INBlock_Device{
public:
    // блокирующий прием
    //*  \return - длинна полученного участка
    virtual int getData ( void* dst, unsigned len) = 0;
    //*  ожидание amount доступных символов
    virtual int get_waitfor(unsigned amount, unsigned to = 0) = 0;
};

class StdIO_Device
  : public StdOUT_Device
{
public:
    virtual int get_char() = 0;
    virtual int get_wait(unsigned to = 0) = 0;
//    virtual int printf ( const char* str) = 0;

private:
};



//-----------------------------------------------------------------------------

// this is lock with value, wich is keep after release, and assigns on lock aquire
// need for some confurable connection sharing

struct ValuedLock{
    public:
    //конструктор добавляет лишнего в структуру
    //ValuedLock(): value(NULL), lock(false) {};

    public:
        typedef void* value_t;

        value_t  value;

        inline
        bool     is_locked() const {return lock;};

        inline
        bool     aquire( value_t x ) {
            if (!lock){
                lock = true;
                value = x;
                return true;
            }
            else
                return false;
        };

        inline
        void     release(){
            lock = false;
        };

    protected:
        bool     lock;
};

// синглетный селектор с флагом блокировки на порт
// позволит произвольно устанавливать точки доступа блокировке на один общий порт.
//  ИСПОЛЬЗОВАНИЕ:
//      class dev1{
//          // доступ к общей блокировке UART1_port
//          PortSingleLock<UARTDevicce, UART1_port> access;
//          ...
//      };
//      class dev2{
//          // доступ к общей блокировке UART1_port
//          PortSingleLock<UARTDevicce, UART1_port> access;
//          ...
//      };
//      void touch_uart1(){
//          // доступ к общей блокировке UART1_port
//          PortSingleLock<UARTDevicce, UART1_port> access;
//
//          //свободный порт хранит свое последнее состояние
//          bool same = (access.value() == uartsel);
//          access.aquire( uartsel );
//          ...
//          access.release();
//      }
template<typename T, T& adress>
class PortSingleLock{
    public:
        //static const u32 port_addr = (u32)(adress);
        typedef T port_t;
        static T&  port() {return adress;};
    public:
        typedef ValuedLock self_t;

        static bool     is_locked() {return single().is_locked();};
        static bool     aquire( self_t::value_t x ) {return single().aquire(x);};
        static void     release(){single().release();};
        static self_t::value_t value() {return single().value;};

    protected:
        static self_t& single(){
            static self_t value;
            return value;
        };
};



// глобальный асинхронный сигнал сделан по мотиву contiki semaphore:
//      глобальный счетчик источника сигнала, и локальный счетчик просмотреного
//      сигнала     - догоняет источник.
//      give()      - сигналит всем обозревателям сигнала
//      try_take()  - true если был новый сигнал, и запоминает что он получен
template<int id>
class GlogalSignal {
    public:

    // забирает имеющийся сигнал
    // \return true - есть непросмотреный глобальный сигнал
    bool try_take(){
        unsigned now = single().stamp;
        if (now == my_stamp)
            return false;
        my_stamp = now;
        return true;
    }

    // сигналит всем обозревателям сигнала
    static bool give(){
        // TODO: сделать атомарно!!!
        single().stamp++;
        return ptOK;
    };

    protected:
        unsigned    my_stamp;
        struct self_t{
            unsigned stamp;
        };

        static self_t& single(){
            static self_t value = {0};
            return value;
        };
};





#endif // INTFACES_H
